/********** Openshop **********/

use io;

/* Read instance data. The input files follow the "Taillard" format */
function input() {
    local usage = "Usage: localsolver openshop.lsp inFileName=instanceFile "
            + "[outFileName=outputFile] [lsTimeLimit=timeLimit]";
    if (inFileName == nil) throw usage;

    inFile = io.openRead(inFileName);
    inFile.readln();
    nbJobs = inFile.readInt();
    nbMachines = inFile.readInt();
    inFile.readln();
    inFile.readln();
    // Processing times for each job on each machine
    // (given in the task order, the processing order is a decision variable)
    processingTimesActivityOrder[j in 0...nbJobs][m in 0...nbMachines] = inFile.readInt();
    inFile.readln();
    // Reorder processing times: processingTime[j][m] is the processing time of the
    // task of job j that is processed on machine m
    for [j in 0...nbJobs][k in 0...nbMachines] {
        local m = inFile.readInt() - 1;
        processingTimes[j][m] = processingTimesActivityOrder[j][k];
    }

    inFile.close();

    maxStart = sum[j in 0...nbJobs][m in 0...nbMachines](processingTimes[j][m]);
}

/* Declare the optimization model */
function model() {
    // Interval decisions: time range of jobs on each machine
    // tasks[j][m] is the interval of time of the task of job j 
    // which is processed on machine m
    tasks[j in 0...nbJobs][m in 0...nbMachines] <- interval(0, maxStart);

    // Task duration constraints
    for [j in 0...nbJobs][m in 0...nbMachines]
        constraint length(tasks[j][m]) == processingTimes[j][m];

    // List of the jobs on each machine
    jobsOrder[m in 0...nbMachines] <- list(nbJobs);
    for [m in 0...nbMachines] {
        // Each job is scheduled on every machine
        constraint count(jobsOrder[m]) == nbJobs;

        // Every machine executes a single task at a time
        constraint and(0...nbJobs - 1, i => tasks[jobsOrder[m][i]][m] < tasks[jobsOrder[m][i + 1]][m]);
    }

    // List of the machines for each job task
    machinesOrder[j in 0...nbJobs] <- list(nbMachines);
    for [j in 0...nbJobs] {
        // Every task is scheduled on its corresponding machine
        constraint count(machinesOrder[j]) == nbMachines;

        // A job has a single task at a time
        constraint and(0...nbMachines - 1, k => tasks[j][machinesOrder[j][k]] < tasks[j][machinesOrder[j][k + 1]]);
    }

    // Minimize the makespan: the end of the last task
    makespan <- max[m in 0...nbMachines][j in 0...nbJobs](end(tasks[j][m]));
    minimize makespan;
}

// Parametrize the solver
function param() {
    if (lsTimeLimit == nil) lsTimeLimit = 5;
}

/* Write the solution in a file with the following format:
 *  - for each machine, the job sequence */
function output() {
    if (outFileName == nil) return;
    outFile = io.openWrite(outFileName);
    for [m in 0...nbMachines] {
        outFile.println[i in 0...nbJobs](jobsOrder[m].value[i] + " ");
    }
    println("Solution written in " + outFileName);
    outFile.close();
}
